/*******************************************************************************
 * Copyright (c) 2014 Michael Simon.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Haykuhi Musheghyan - 2025
 ******************************************************************************/

package edu.kit.scc.webreg.bean;

import java.io.IOException;
import java.io.Serializable;
import java.security.PrivateKey;
import java.util.Calendar;
import java.util.Date;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWSAlgorithm;
import com.nimbusds.jose.JWSHeader;
import com.nimbusds.jose.crypto.RSASSASigner;
import com.nimbusds.jwt.JWT;
import com.nimbusds.jwt.JWTClaimsSet;
import com.nimbusds.jwt.JWTParser;
import com.nimbusds.jwt.SignedJWT;
import com.nimbusds.oauth2.sdk.Scope;
import com.nimbusds.oauth2.sdk.token.BearerAccessToken;
import com.nimbusds.openid.connect.sdk.token.OIDCTokens;

import edu.kit.scc.webreg.entity.identity.IdentityEntity;
import edu.kit.scc.webreg.entity.oidc.OidcOpConfigurationEntity;
import edu.kit.scc.webreg.service.identity.IdentityService;
import edu.kit.scc.webreg.service.oidc.OidcOpConfigurationService;
import edu.kit.scc.webreg.service.saml.CryptoHelper;
import edu.kit.scc.webreg.service.saml.exc.OidcAuthenticationException;
import edu.kit.scc.webreg.session.SessionManager;
import edu.kit.scc.webreg.util.FacesMessageGenerator;
import edu.kit.scc.webreg.util.ResourceBundleHelper;
import jakarta.faces.view.ViewScoped;
import jakarta.inject.Inject;
import jakarta.inject.Named;
import jakarta.servlet.http.HttpServletRequest;
import lombok.Getter;
import lombok.Setter;

@Named
@ViewScoped
public class CreateAccessTokensBean implements Serializable {
	private static final long serialVersionUID = 1L;

	@Inject
	private SessionManager session;

	@Inject
	private HttpServletRequest request;

	@Inject
	private IdentityService identityService;

	@Inject
	private FacesMessageGenerator messageGenerator;

	@Inject
	private OidcOpConfigurationService oidcOpConfigService;

	@Inject
	private CryptoHelper cryptoHelper;

	@Inject
	private ResourceBundleHelper resourceHelper;

	private IdentityEntity identity;

	@Getter
	@Setter
	private Date validUntil;

	@Getter
	@Setter
	private OidcOpConfigurationEntity opConfig;

	@Getter
	@Setter
	private String generatedAccessToken;

	@Getter
	@Setter
	private String tokenValidationMsg;

	@Getter
	@Setter
	private String tokenToValidate;
	
	@Getter
	@Setter
	private String scope;
	
	
	private Date today;
	
    private Date maxDate;


    public Date getToday() {
    	today = new Date(); 
    	return today;
    }
	
    public Date getMaxDate() {
       	
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(getToday());
        calendar.add(Calendar.DATE, 14);
        maxDate = calendar.getTime();
        return maxDate;
    }
    
	public IdentityEntity getIdentity() {
		if (identity == null) {
			identity = identityService.fetch(session.getIdentityId());
		}
		return identity;

	}

	public boolean isDateValid(Date dt) {

		if (dt.before(new Date()))
			return false;
		return true;
	}

	public void generateAnAccessToken() throws OidcAuthenticationException, IOException {

		if (!isDateValid(getValidUntil())) {
			messageGenerator.addResolvedErrorMessage("invalid_date", "invalid_date_details", true);
			return;
		}

		opConfig = oidcOpConfigService.findByAttr("host", request.getServerName());
		
		if (opConfig == null) {
			messageGenerator.addResolvedErrorMessage("failure", "load_config_details", true);
			return;
		}
		
		
		scope = getScope();
		
		if(scope.isEmpty()) {
			messageGenerator.addResolvedErrorMessage("failure", "Select the token scope", false);
			return;
		}
				
		if (getIdentity() != null && (scope.equals("Project") || scope.equals("Login"))) {
			
	        String audienceUrl = (scope.equals("Project")) 
	            ? "https://" + opConfig.getHost() + "/rest/project/" 
	            : "https://" + opConfig.getHost() + "/login";
	        
	        
	        JWTClaimsSet.Builder claimsBuilder = new JWTClaimsSet.Builder();
	        claimsBuilder
	                .issuer("https://" + opConfig.getHost() + "/oidc/realms/" + opConfig.getRealm())
	                .issueTime(new Date())
	                .expirationTime(getValidUntil())
	                .subject(session.getIdentityId().toString())
	                .audience(audienceUrl);
	        
	        JWTClaimsSet claims = claimsBuilder.build();

			SignedJWT jwt;
			try {
				PrivateKey privateKey = cryptoHelper.getPrivateKey(opConfig.getPrivateKey());
				RSASSASigner rsaSigner = new RSASSASigner(privateKey);
				JWSHeader header = new JWSHeader.Builder(JWSAlgorithm.RS256).build();
				jwt = new SignedJWT(header, claims);
				jwt.sign(rsaSigner);

				setGeneratedAccessToken(jwt.serialize());

			} catch (JOSEException e) {
				throw new OidcAuthenticationException(e);
			}

			//BearerAccessToken bat = new BearerAccessToken(3600, new Scope(opConfig.getHost()));
			//OIDCTokens tokens = new OIDCTokens(jwt, bat, null);
		}
	}

	public void validateAccessToken(String token) {

		try {

			JWT jwt = JWTParser.parse(token);
			JWTClaimsSet claims = jwt.getJWTClaimsSet();

			if (!claims.getIssuer().contains(request.getServerName())) {
				String details = resourceHelper.resolveMessage("token_issuer_details");
				setTokenValidationMsg(details + " " + claims.getIssuer());
			}
			Date expirationTime = claims.getExpirationTime();
			    if (expirationTime != null && expirationTime.before(new Date())) {			    	
			        String details = resourceHelper.resolveMessage("token_expired_details");
			        setTokenValidationMsg(details + " " + expirationTime);
			    } else {
			        String details = resourceHelper.resolveMessage("token_valid_until_details");
			        setTokenValidationMsg(details + " " + expirationTime);
			    }
			

		} catch (Exception e) {

			String details = resourceHelper.resolveMessage("invalid_token_details");
			setTokenValidationMsg(details);
		}

	}

}
